# -*- coding: utf-8 -*-
"""
Created on Mon Feb 25 18:23:15 2019

@author: Kathy_Breen
"""

from keras.layers import Input, Dense, Concatenate, Dropout, BatchNormalization, CuDNNLSTM
from keras.models import Model
from keras.callbacks import EarlyStopping, ReduceLROnPlateau
from keras.initializers import Ones
from keras.optimizers import Adam, Nadam#, RMSprop
import h5py
import numpy as np
#import itertools
import pandas as pd
import os
import pickle as pkl

##%% SETTINGS FOR REPRODUCIBLE RESULTS DURING DEVELOPMENT
#
##import numpy as np
#import tensorflow as tf
#import random as rn
#
## The below is necessary in Python 3.2.3 onwards to
## have reproducible behavior for certain hash-based operations.
## See these references for further details:
## https://docs.python.org/3.4/using/cmdline.html#envvar-PYTHONHASHSEED
## https://github.com/keras-team/keras/issues/2280#issuecomment-306959926
#
##import os
#os.environ['PYTHONHASHSEED'] = '0'
#
## The below is necessary for starting Numpy generated random numbers
## in a well-defined initial state.
#
#np.random.seed(42)
#
## The below is necessary for starting core Python generated random numbers
## in a well-defined state.
#
#rn.seed(12345)
#
## Force TensorFlow to use single thread.
## Multiple threads are a potential source of
## non-reproducible results.
## For further details, see: https://stackoverflow.com/questions/42022950/which-seeds-have-to-be-set-where-to-realize-100-reproducibility-of-training-res
#
#session_conf = tf.compat.v1.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)
#
#from keras import backend as K
#
## The below tf.set_random_seed() will make random number generation
## in the TensorFlow backend have a well-defined initial state.
## For further details, see: https://www.tensorflow.org/api_docs/python/tf/set_random_seed
#
##tf.set_random_seed(1234)
#tf.compat.v1.set_random_seed(1234)
#
#sess = tf.compat.v1.Session(graph=tf.compat.v1.get_default_graph(), config=session_conf)
#K.set_session(sess)

#%% WORK
from keras import backend as K
import tensorflow as tf

# define custom loss function

def sameSign(x,y):
    # returns 0 if x and y are of opposite sign, and 1 if the signs are the same
    xsign = tf.sign(x)
    ysign = tf.sign(y)
    return tf.to_float(tf.equal(xsign,ysign))
    

def pblf(ET_DATA,ET_SHIFT):
    
    # ET_DATA is an array the same size as y_true with daily ET values temporally sync all samples in Y
    # Y_SHIFT is an array the same size as y_true where y_true[i,:] == Y_SHIFT[i-1,:]
    # ET_SHIFT is an array the same size as ET_DATA where ET_DATA[i,:] == ET_SHIFT[i-1,:]
    
    def loss(y_true,y_pred):
    
        Y_SHIFT = tf.roll(y_pred,1,axis=0)

        y_pred_diff = tf.subtract(y_pred,Y_SHIFT)
        et_diff = tf.subtract(ET_DATA,ET_SHIFT)
        
        # test if soil moisture decreased when et increased and v.v. - penalize loss function if physical relationship does not hold, i.e. the signs are the same
        delta_sign = sameSign(y_pred_diff,et_diff)
        return K.mean(K.square(y_pred-y_true) + 0.1*delta_sign*K.sqrt(K.square(et_diff) + K.square(y_pred_diff)),axis=-1)
    
    return loss

#%%

flag = ''
#flag = '_noflag'

with h5py.File('X_D_LSTM'+flag+'.hdf5','r') as f:
    XD_train = np.array(f["XD_train"])
    XD_test = np.array(f["XD_test"])
        
with h5py.File('X_S_LSTM.hdf5','r') as f:
    XS_train = np.array(f["XS_train"])
    XS_test = np.array(f["XS_test"])
        
with h5py.File('Y_LSTM.hdf5','r') as f:
    Ysmap_train = np.array(f["Y_train"])
    Ysmap_test = np.array(f["Y_test"])
    
with h5py.File('Y_ET_LSTM.hdf5','r') as f:
    Yet_train = np.array(f["Yet_train"])
    Yet_test = np.array(f["Yet_test"])
    Yet_shift_train = np.array(f["Yet_shift_train"])
    Yet_shift_test = np.array(f["Yet_shift_test"])
        
with h5py.File('Test0'+flag+'.hdf5','r') as f:
    XD_test0 = np.array(f['XD_test0'])
    XS_test0 = np.array(f['XS_test0'])
    Ysmap_test0 = np.array(f['Y_test0'])
    Yet_test0 = np.array(f['Yet_test0'])
    Yet_shift_test0 = np.array(f["Yet_shift_test0"])

# the model is trained on 1096 timesteps - 01/01/2016-12/31/2018. the input arrays have data from 01/01/16-12/31/19 the output arrays all have data from 3/31/15-12/31/19 - limit the data I/O to the time period for training
    
XD_train = XD_train[:,:1096,:]
XD_test = XD_test[:,:1096,:]
XD_test0 = XD_test0[:,:1096,:]
    
Ysmap_train = Ysmap_train[:,276:276+1096]
Yet_train = Yet_train[:,276:276+1096]
Yet_shift_train = Yet_shift_train[:,276:276+1096]

Ysmap_test = Ysmap_test[:,276:276+1096]
Yet_test = Yet_test[:,276:276+1096]
Yet_shift_test = Yet_shift_test[:,276:276+1096]

Ysmap_test0 = Ysmap_test0[:,276:276+1096]
Yet_test0 = Yet_test0[:,276:276+1096]
Yet_shift_test0 = Yet_shift_test0[:,276:276+1096]

padding_value = -1.0

XD_Ntimesteps = XD_train.shape[1]
XD_Nfeatures = XD_train.shape[2]
XS_Nfeatures = XS_train.shape[1]
Ysmap_Nfeatures = Ysmap_train.shape[1]

epochs = 500

#%% define functions

def LSTM_PGNN_addtnl_params(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_test,params):

    main_input = Input(shape=(XD_Ntimesteps,XD_Nfeatures),
                   dtype='float',
                   name='Dynamic_Input'
                   )
    x = BatchNormalization()(main_input)
       
    if params['Nlyr_LSTM']>2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        for i in range(1,params['Nlyr_LSTM']-1):
            lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
        
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    elif params['Nlyr_LSTM'] == 2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    else:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
            
    auxiliary_output = Dense(Ysmap_Nfeatures, name='aux_output', activation=params['MLP_activation'])(lstm_out)
    auxiliary_input = Input(shape=(XS_Nfeatures,)) 
    x = Concatenate(axis=-1)([lstm_out, auxiliary_input]) 
    
    x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
    x = BatchNormalization()(x)
    Dropout(params['do'])(x)
        
    if params['Nlyr_MLP'] > 1:
        for i in range(1,params['Nlyr_MLP']):
            x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
            x = BatchNormalization()(x)
            Dropout(params['do'])(x)
    
    final_output = Dense(Ysmap_Nfeatures, name='final_output', activation=params['MLP_activation'])(x)
    
    # create the placeholder tensors for the PBLF function
    y_et = Input(shape=(Yet_train.shape[1],), name='y_et') 
    et_shift = Input(shape=(Yet_train.shape[1],), name='et_shift')  
    
    # define model
    model = Model(inputs=[main_input, auxiliary_input, y_et, et_shift], 
                  outputs=[final_output, auxiliary_output]
                  )
    
#    adagrad = Adagrad(clipvalue=0.01)  # employ gradient clipping to avoid nans in loss fxn
    opt = params['optimizer']
    if params['TF_pblf'] ==1:
        model.compile(loss=pblf(ET_DATA=y_et,ET_SHIFT=et_shift),
                      optimizer=opt,
                      metrics=['mse'])
    else:
        model.compile(loss='mean_squared_error',
                      optimizer=opt,
                      metrics=['mse'])
    
    reduce_lr = ReduceLROnPlateau(monitor=params['monitor'],
                                  factor=params[ 'lr_factor'],  
                                  patience=params['patience_es']-5,
                                  verbose=1)
    
    early_stop = EarlyStopping(monitor=params['monitor'], 
                               min_delta=params['min_delta'], 
                               patience=params['patience_es'], 
                               verbose=1)  
    
    #print('Train the model............')
    if params['TFlr'] == 1:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
#                callbacks=[reduce_lr],
                callbacks=[reduce_lr,early_stop],
                verbose=2
                        )
    else:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train,],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
                callbacks=[early_stop],
                verbose=2
                )
    
    return history, model
    
 
#%% define hyperparameter combinations
# Every couple of days during HP tuning, the output of the HP tuning loop was examined and the best performers were selected. Their configs are below..they are run in a loop here with no seed set    

#param_lst = [[5],  # Nnodes_LSTM
#             [10],  # Nnodes_MLP
#             [1,2],  # Nlyr_LSTM
#             [1,10],  # Nlyr_MLP
#             [75,100],  # batch_size
#             ['relu','elu'],  # MLP_activation
#             [RMSprop(),Adam(),Nadam()],  # optimizer
#             [0.1,0.2,0.3]  # dropout rate
#             ]

# best performers from HP tuning runs as of 3/30/2020
#param_tuples = [
#        (5,10,1,1,75,'relu',RMSprop(),0.2),
#        (5,10,1,1,75,'relu',RMSprop(),0.3),
#        (5,10,1,1,75,'relu',RMSprop(),0.1),
#        (5,10,1,1,75,'elu',RMSprop(),0.1),
#        (5,10,1,1,75,'elu',RMSprop(),0.3),
#        (5,10,1,10,100,'elu',Adam(),0.2),
#        (5,10,1,10,100,'elu',Nadam(),0.1),
#        (5,10,1,10,100,'elu',Nadam(),0.3),
#        (5,10,2,1,100,'relu',Nadam(),0.1),
#        (5,10,1,1,75,'relu',Nadam(),0.2)
#        ]

## 4/1/2020
#param_tuples = [
#    (5,100,1,10,100,'elu',Nadam(),0.2),
#    (10,10,1,2,75,'elu',Adam(),0.3),
#    (10,100,1,3,75,'elu',Adam(),0.1),
#    (5,100,1,10,100,'relu',Nadam(),0.3),
#    (10,100,1,3,75,'elu',Adam(),0.2),
#    (10,10,1,3,100,'elu',RMSprop(),0.1),
#    (5,50,1,2,75,'elu',Nadam(),0.3),
#    (10,10,1,3,100,'relu',RMSprop(),0.3),
#    (5,10,1,10,100,'elu',Adam(),0.2),
#    (10,10,1,2,100,'elu',Nadam(),0.1)
#]

## 4/3/2020
#param_tuples = [
#    (10,10,1,3,75,'elu',Adam(),0.3),
#    (5,10,2,1,100,'elu',Nadam(),0.2),
#    (5,10,1,1,75,'elu',Nadam(),0.2),
#    (5,10,1,2,100,'elu',RMSprop(),0.2),
#    (5,100,1,10,100,'elu',Nadam(),0.2),
#    (5,50,1,3,100,'elu',Nadam(),0.1),
#    (10,10,1,3,75,'elu',Adam(),0.2),
#    (5,10,1,1,100,'elu',Nadam(),0.2),
#    (5,10,1,10,100,'elu',RMSprop(),0.1),
#    (10,10,1,3,75,'elu',Nadam(),0.2)
#    ]

## 4/6/2020
#param_tuples = [
#    (5,100,1,10,100,'elu',Nadam(),0.2),
#    (10,10,1,2,75,'elu',Adam(),0.3),
#    (5,50,1,3,100,'elu',Nadam(),0.1),
#    (10,100,1,3,75,'elu',Adam(),0.1),
#    (5,100,1,10,100,'relu',Nadam(),0.3),
#    (10,100,1,3,75,'elu',Adam(),0.2),
#    (10,10,1,3,100,'elu',RMSprop(),0.1),
#    (5,50,1,2,75,'elu',Nadam(),0.3),
#    (10,10,1,3,100,'relu',RMSprop(),0.3),
#    (5,10,1,10,100,'elu',Adam(),0.2),
#    (10,10,1,3,75,'elu',Adam(),0.3),
#    (5,10,2,1,100,'elu',Nadam(),0.2),
#    (5,10,1,1,75,'elu',Nadam(),0.2),
#    (5,10,1,2,100,'elu',RMSprop(),0.2),
#    (10,10,1,3,75,'elu',Adam(),0.2),
#    (5,10,1,1,100,'elu',Nadam(),0.2),
#    (5,10,1,10,100,'elu',RMSprop(),0.1),
#    (10,10,1,3,75,'elu',Nadam(),0.2)
#]

## 4/8/2020 using additional optimization parameters on 4/6 best performers
#param_tuples = [
#(5,100,1,10,100,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_mse',1),
#(5,100,1,10,100,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_mse',1),
#(5,100,1,10,100,'elu',Nadam(),0.2,0.001,1,20,0.8,'val_final_output_loss',1),
#(5,50,1,3,100,'elu',Nadam(),0.1,0,1,20,0.5,'val_final_output_mse',1),
#(5,50,1,2,75,'elu',Nadam(),0.3,0,1,20,0.5,'val_final_output_mse',1),
#(10,100,1,3,75,'elu',Adam(),0.1,0,1,20,0.5,'val_final_output_mse',1),
#(5,10,1,10,100,'elu',Adam(),0.2,0,1,20,0.5,'val_final_output_mse',1),
#(10,10,1,3,100,'elu',RMSprop(),0.1,0,1,20,0.5,'val_final_output_mse',1),
#(10,10,1,3,75,'elu',Adam(),0.3,0,1,20,0.5,'val_final_output_mse',1),
#(10,10,1,2,75,'elu',Adam(),0.3,0.001,1,20,0.8,'val_final_output_loss',1),
#(5,10,2,1,100,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_loss',0),
#(5,100,1,10,100,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_mse',0),
#(10,10,1,2,75,'elu',Adam(),0.3,0,1,20,0.5,'val_final_output_mse',0),
#(10,10,1,3,100,'relu',RMSprop(),0.3,0,1,20,0.5,'val_final_output_mse',0),
#(5,100,1,10,100,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_mse',0),
#(5,10,1,1,75,'elu',Nadam(),0.2,0,1,20,0.5,'val_final_output_mse',0),
#(5,100,1,10,100,'relu',Nadam(),0.3,0,1,20,0.5,'val_final_output_mse',0),
#(5,100,1,10,100,'relu',Nadam(),0.3,0,1,20,0.75,'val_final_output_loss',0),
#(5,100,1,10,100,'relu',Nadam(),0.3,0,1,100,0.8,'val_final_output_mse',0),
#(5,100,1,10,100,'relu',Nadam(),0.3,0,1,50,0.75,'val_final_output_loss',0)
#]

# 4/29/2020
#param_tuples = [
##(5,100,1,10,100,'elu',Nadam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_mse',1),
##(5,100,1,10,100,'elu',Nadam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,50,1,3,100,'elu',Nadam(clipvalue=0.01),0.1,0,1,20,0.5,'val_final_output_mse',1),
##(5,50,1,2,75,'elu',Nadam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',1),
##(10,100,1,3,75,'elu',Adam(clipvalue=0.01),0.1,0,1,20,0.5,'val_final_output_mse',1),
##(5,10,1,10,100,'elu',Adam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_mse',1),
##(10,10,1,3,100,'elu',RMSprop(clipvalue=0.01),0.1,0,1,20,0.5,'val_final_output_mse',1),
##(10,10,1,3,75,'elu',Adam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',1),
##(10,10,1,2,75,'elu',Adam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,50,1,3,100,'elu',Nadam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,75,'elu',Adam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_mse',1),
##(10,100,1,3,75,'elu',Adam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,75,'elu',Adam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,10,1,3,100,'elu',RMSprop(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,50,1,2,75,'elu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,10,1,3,100,'relu',RMSprop(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,10,1,2,75,'elu',Adam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',1),
##(5,10,1,10,100,'elu',Adam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,10,1,2,100,'elu',Nadam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_loss',1),
##(10,100,1,3,100,'relu',Nadam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,100,'relu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,100,1,10,100,'elu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
#(5,100,1,10,75,'elu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),  # BP PBLF
##(5,100,1,10,100,'elu',Adam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,10,1,10,100,'elu',Nadam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,100,'relu',Nadam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,75,'relu',Adam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,50,1,2,75,'relu',Nadam(clipvalue=0.01),0.2,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,50,1,3,75,'relu',Nadam(clipvalue=0.01),0.1,0.0001,1,20,0.8,'val_final_output_loss',1),
##(10,100,1,3,100,'elu',Nadam(clipvalue=0.01),0.3,0.0001,1,20,0.8,'val_final_output_loss',1),
##(5,10,2,1,100,'elu',Nadam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_loss',0),
##(10,10,1,2,75,'elu',Adam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',0),
##(10,10,1,3,100,'relu',RMSprop(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'elu',Nadam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_mse',0),
##(5,10,1,1,75,'elu',Nadam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,0,100,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,20,0.75,'val_final_output_loss',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0.001,0,50,0.8,'val_final_output_loss',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,100,0.8,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,0,100,0.8,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0.001,1,20,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,50,0.75,'val_final_output_loss',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,20,0.5,'val_final_output_loss',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,100,0.5,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,20,0.75,'val_final_output_mse',0),
#(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,0,100,0.8,'val_final_output_loss',0),  # BP MSE
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,0,50,0.8,'val_final_output_loss',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,0,50,0.8,'val_final_output_mse',0),
##(5,100,1,10,100,'relu',Nadam(clipvalue=0.01),0.3,0,1,50,0.5,'val_final_output_loss',0)
#]

# top three performers each for configs optimizinf MSE and PBLF
#param_tuples = [
#(5,10,2,1,100,'elu',Nadam(clipvalue=0.01),0.2,0,1,20,0.5,'val_final_output_loss',0),
#(10,10,1,2,75,'elu',Adam(clipvalue=0.01),0.3,0,	1,	20,	0.5,	'val_final_output_mse',	0),
#(10,	10,	1,	3,	100,	'relu',	RMSprop(clipvalue=0.01),	0.3,	0,	1,	20,	0.5,	'val_final_output_mse',	0),
#(5,	100,	1,	10,	100,	'elu',	Nadam(clipvalue=0.01),	0.2,	0,1,	20,	0.5,	'val_final_output_mse',	1),
#(5,	100,	1,	10,	100,	'elu',	Nadam(clipvalue=0.01),	0.2,	0.0001,	1,	20,	0.8,	'val_final_loss',	1),
#(5,	50,	1,	3,	100,	'elu',	Nadam(clipvalue=0.01),	0.1,	0,	1,	20,	0.5,	'val_final_output_mse',	1)
#]

#param_tuples = [
#(1,	150,	1,50,	50,	'elu',	Nadam(clipvalue=0.01),	0.3,	0.0001,	1,	250,	0.8,	'val_final_output_loss',	1),
#(1,	150,	1,	50,	50,	'elu',	Nadam(clipvalue=0.01),	0.3,	0.0001,	0,	250,	0.8,	'val_final_output_loss',	0),
#(2,	150,	1,	50,	50,	'relu',	Nadam(clipvalue=0.01),	0.3,	0.0001,	0,	250,	0.8,	'val_final_output_loss',	1),
#(2,	150,	1,	50,	50,	'elu',	Nadam(clipvalue=0.01),	0.3,	0,	0,	250,	0.8,	'val_final_output_loss',	0),
#(3,	150,	1,	50,	50,	'elu'	,Nadam(clipvalue=0.01),	0.3,	0.0001,	0,	250,	0.8,	'val_final_output_loss',	1),
#(3,	150,	1,	50,	50,	'elu',	Adam(clipvalue=0.01),	0.2,	0.0001,	1,	250,	0.8,	'val_final_output_loss',	1),
#]

# the best-performing model is stored on SCJs machine...I don't have access to it right now so retrain it here
#param_tuples = [(5,	100,	1,	10,	100,	'elu',	Nadam(clipvalue=0.01),	0.2,	0.0001,	1,	20,	0.8,	'val_final_loss',	1)]
param_tuples = [(3,	150,	1,	50,	50,	'elu',	Adam(clipvalue=0.01),	0.3,	0,	0,	250,	0.8,	'val_final_loss',	0)]




#%% Call the function to perform hyperparaeter tuning

for i,tup in enumerate(param_tuples):

        
    params = {
    'Model_name': 'BP_model_'+str(i),
    'Nnodes_LSTM': tup[0],
    'Nnodes_MLP': tup[1],
    'Nlyr_LSTM': tup[2],
    'Nlyr_MLP': tup[3],
    'batch_size': tup[4],
    'MLP_activation': tup[5],
    'optimizer': tup[6],
    'do': tup[7],
    'min_delta': tup[8],  # min delta for early stop
    'TFlr': tup[9],  # true or false to use learning rate reduction (some optimizers do this anyway)
    'patience_es': tup[10],  # patience for early stopping
    'lr_factor': tup[11],  # factor for lr reduction
    'monitor': tup[12],  # quantity to monitor for callbacks
    'TF_pblf': tup[13]  # use pblf or mse
    }
    
    history, model = LSTM_PGNN_addtnl_params(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_test,params)
    
    histdict = history.history
    pkl.dump(histdict, open('BP_histdict_'+str(i)+'_'+flag+'.pkl', "wb" ))
    model.save('BP_model_'+str(i)+'_'++flag+'.h5')
        
    params['optimizer'] = str(params['optimizer'])
    params['final_output_loss'] = histdict['val_final_output_loss'][len(histdict['val_final_output_loss'])-1]
    params['final_output_mse'] = histdict['val_final_output_mse'][len(histdict['val_final_output_mse'])-1]
    
    param_df = pd.DataFrame(params,index=[0])

    if not os.path.exists('BPout_extra_runs.csv'):
        param_df.to_csv('BPout_extra_runs.csv', mode='w+', index=False, header=True)
    else:
        param_df.to_csv('BPout_extra_runs.csv', mode='a', index=False, header=False)













































































