# -*- coding: utf-8 -*-
"""
Created on Mon Feb 25 18:23:15 2019

@author: Kathy_Breen
"""

from keras.layers import Input, Dense, Concatenate, Dropout, BatchNormalization, CuDNNLSTM
from keras.models import Model
from keras.callbacks import EarlyStopping, ReduceLROnPlateau
from keras.initializers import Ones
from keras.optimizers import Adam, Nadam, RMSprop
#from keras.optimizers import Adadelta, Adamax, Adam, Nadam
import h5py
import numpy as np
import itertools
import pandas as pd
import os


#%% WORK
from keras import backend as K
import tensorflow as tf

# define custom loss function

def sameSign(x,y):
    # returns 0 if x and y are of opposite sign, and 1 if the signs are the same
    xsign = tf.sign(x)
    ysign = tf.sign(y)
    return tf.to_float(tf.equal(xsign,ysign))
    

def pblf(ET_DATA,ET_SHIFT):
    
    # ET_DATA is an array the same size as y_true with daily ET values temporally sync all samples in Y
    # Y_SHIFT is an array the same size as y_true where y_true[i,:] == Y_SHIFT[i-1,:]
    # ET_SHIFT is an array the same size as ET_DATA where ET_DATA[i,:] == ET_SHIFT[i-1,:]
    
    def loss(y_true,y_pred):
    
        Y_SHIFT = tf.roll(y_pred,1,axis=0)

        y_pred_diff = tf.subtract(y_pred,Y_SHIFT)
        et_diff = tf.subtract(ET_DATA,ET_SHIFT)
        
        # test if soil moisture decreased when et increased and v.v. - penalize loss function if physical relationship does not hold, i.e. the signs are the same
        delta_sign = sameSign(y_pred_diff,et_diff)
        return K.mean(K.square(y_pred-y_true) + 0.1*delta_sign*K.sqrt(K.square(et_diff) + K.square(y_pred_diff)),axis=-1)
    
    return loss

#%%


with h5py.File('X_D_LSTM.hdf5','r') as f:
    XD_train = np.array(f["XD_train"])
    XD_test = np.array(f["XD_test"])
        
with h5py.File('X_S_LSTM.hdf5','r') as f:
    XS_train = np.array(f["XS_train"])
    XS_test = np.array(f["XS_test"])
        
with h5py.File('Y_LSTM.hdf5','r') as f:
    Ysmap_train = np.array(f["Y_train"])
    Ysmap_test = np.array(f["Y_test"])
    
with h5py.File('Y_ET_LSTM.hdf5','r') as f:
    Yet_train = np.array(f["Yet_train"])
    Yet_test = np.array(f["Yet_test"])
    Yet_shift_train = np.array(f["Yet_shift_train"])
    Yet_shift_test = np.array(f["Yet_shift_test"])
        
with h5py.File('Test0.hdf5','r') as f:
    XD_test0 = np.array(f['XD_test0'])
    XS_test0 = np.array(f['XS_test0'])
    Ysmap_test0 = np.array(f['Y_test0'])
    Yet_test0 = np.array(f['Yet_test0'])
    Yet_shift_test0 = np.array(f["Yet_shift_test0"])


Ysmap_train = Ysmap_train.reshape(Ysmap_train.shape[0],Ysmap_train.shape[1])
Ysmap_test = Ysmap_test.reshape(Ysmap_test.shape[0],Ysmap_test.shape[1])
Ysmap_test0 = Ysmap_test0.reshape(Ysmap_test0.shape[0],Ysmap_test0.shape[1])

padding_value = -1.0

XD_Ntimesteps = XD_train.shape[1]
XD_Nfeatures = XD_train.shape[2]
XS_Nfeatures = XS_train.shape[1]
Ysmap_Nfeatures = Ysmap_train.shape[1]

epochs = 250

#%% define function
def LSTM_PGNN(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_shift_test,Ysmap_test,params):

    main_input = Input(shape=(XD_Ntimesteps,XD_Nfeatures),
                   dtype='float',
                   name='Dynamic_Input'
                   )
    x = BatchNormalization()(main_input)
       
    if params['Nlyr_LSTM']>2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        for i in range(1,params['Nlyr_LSTM']-1):
            lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
        
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    elif params['Nlyr_LSTM'] == 2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    else:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
            
    auxiliary_output = Dense(Ysmap_Nfeatures, name='aux_output', activation=params['MLP_activation'])(lstm_out)
    auxiliary_input = Input(shape=(XS_Nfeatures,)) 
    x = Concatenate(axis=-1)([lstm_out, auxiliary_input]) 
    
    x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
    x = BatchNormalization()(x)
    Dropout(params['do'])(x)
        
    if params['Nlyr_MLP'] > 1:
        for i in range(1,params['Nlyr_MLP']):
            x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
            x = BatchNormalization()(x)
            Dropout(params['do'])(x)
    
    final_output = Dense(Ysmap_Nfeatures, name='final_output', activation=params['MLP_activation'])(x)
    y_et = Input(shape=(Yet_train.shape[1],), name='y_et') 
    et_shift = Input(shape=(Yet_train.shape[1],), name='et_shift') 
    y_shift = Input(shape=(Ysmap_shift_train.shape[1],), name='y_shift') 
    
    # define model
    model = Model(inputs=[main_input, auxiliary_input, y_et, et_shift, y_shift], 
                  outputs=[final_output, auxiliary_output]
                  )
    
#    adagrad = Adagrad(clipvalue=0.01)  # employ gradient clipping to avoid nans in loss fxn
    opt = params['optimizer']
    model.compile(loss=pblf(ET_DATA=y_et,ET_SHIFT=et_shift),
                  #loss='mean_squared_error',
                  optimizer=opt,
                  metrics=['mse'])
    
    reduce_lr = ReduceLROnPlateau(monitor='val_final_output_loss',
                                  factor=0.8,  
                                  patience=15,
                                  verbose=1)
    
    early_stop = EarlyStopping(monitor='val_final_output_loss', 
                               min_delta=0.0001,  
                               patience=20, 
                               verbose=1)  
    
    #print('Train the model............')
    history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train,],[Ysmap_train,Ysmap_train],
                    epochs=epochs,
                    batch_size=params['batch_size'],
                    validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
                    callbacks=[reduce_lr,early_stop],
                    verbose=2
                    )
    
    return history, model
    
# OPTIMIZE ON ADDITIONAL PARAMS
def LSTM_PGNN_addtnl_params(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_shift_test,Ysmap_test,params):

    main_input = Input(shape=(XD_Ntimesteps,XD_Nfeatures),
                   dtype='float',
                   name='Dynamic_Input'
                   )
    x = BatchNormalization()(main_input)
       
    if params['Nlyr_LSTM']>2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        for i in range(1,params['Nlyr_LSTM']-1):
            lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
        
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    elif params['Nlyr_LSTM'] == 2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    else:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
            
    auxiliary_output = Dense(Ysmap_Nfeatures, name='aux_output', activation=params['MLP_activation'])(lstm_out)
    auxiliary_input = Input(shape=(XS_Nfeatures,)) 
    x = Concatenate(axis=-1)([lstm_out, auxiliary_input]) 
    
    x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
    x = BatchNormalization()(x)
    Dropout(params['do'])(x)
        
    if params['Nlyr_MLP'] > 1:
        for i in range(1,params['Nlyr_MLP']):
            x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
            x = BatchNormalization()(x)
            Dropout(params['do'])(x)
    
    final_output = Dense(Ysmap_Nfeatures, name='final_output', activation=params['MLP_activation'])(x)
    y_et = Input(shape=(Yet_train.shape[1],), name='y_et') 
    et_shift = Input(shape=(Yet_train.shape[1],), name='et_shift') 
    
    # define model
    model = Model(inputs=[main_input, auxiliary_input, y_et, et_shift], 
                  outputs=[final_output, auxiliary_output]
                  )
    
#    adagrad = Adagrad(clipvalue=0.01)  # employ gradient clipping to avoid nans in loss fxn
    opt = params['optimizer']
    if params['TF_pblf'] ==1:
        model.compile(loss=pblf(ET_DATA=y_et,ET_SHIFT=et_shift),
                      optimizer=opt,
                      metrics=['mse'])
    else:
        model.compile(loss='mean_squared_error',
                      optimizer=opt,
                      metrics=['mse'])
    
    reduce_lr = ReduceLROnPlateau(monitor=params['monitor'],
                                  factor=params[ 'lr_factor'],  
                                  patience=params['patience_es']-5,
                                  verbose=1)
    
    early_stop = EarlyStopping(monitor=params['monitor'], 
                               min_delta=params['min_delta'], 
                               patience=params['patience_es'], 
                               verbose=1)  
    
    #print('Train the model............')
    if params['TFlr'] == 1:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
                callbacks=[reduce_lr,early_stop],
                verbose=2
                        )
    else:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
                callbacks=[early_stop],
                verbose=2
                )
    
    return history, model
    

#%% define hyperparameter combinations
#param_lst = [[5,10],  # Nnodes_LSTM
#             [10,50,100],  # Nnodes_MLP
#             [1,2],  # Nlyr_LSTM
#             [1,2,3,5,10],  # Nlyr_MLP
#             [75,100,250],  # batch_size
#             ['relu','elu'],  # MLP_activation
#             [Adadelta(), Adamax(), Adam(), Nadam()],  # optimizer
#             [0.1,0.2,0.3]  # dropout rate
#             ]
#
#param_tuples = list(itertools.product(*param_lst))
#
## last iteration, the GPU quit after N runs...pick up where you left off
#param_tuples[:1325] = []
#
### kodiak edit 
##param_tuples = param_tuples[1250:1500]  
    
# now take the best performers from all runs and optimize on additional parameters
# 4/6/2020
#best_tuples = [
##    (5,100,1,10,100,'elu',Nadam(),0.2),  # v21
##    (10,10,1,2,75,'elu',Adam(),0.3),
##    (5,50,1,3,100,'elu',Nadam(),0.1),
##    (10,100,1,3,75,'elu',Adam(),0.1),
##    (5,100,1,10,100,'relu',Nadam(),0.3),
##    (10,100,1,3,75,'elu',Adam(),0.2),
##    (10,10,1,3,100,'elu',RMSprop(),0.1),
##    (5,50,1,2,75,'elu',Nadam(),0.3),
##    (10,10,1,3,100,'relu',RMSprop(),0.3),
##    (5,10,1,10,100,'elu',Adam(),0.2),
##    (10,10,1,3,75,'elu',Adam(),0.3),
##    (5,10,2,1,100,'elu',Nadam(),0.2),
##    (5,10,1,1,75,'elu',Nadam(),0.2),
#    (5,10,1,2,100,'elu',RMSprop(),0.2),
#    (10,10,1,3,75,'elu',Adam(),0.2),
#    (5,10,1,1,100,'elu',Nadam(),0.2),
#    (5,10,1,10,100,'elu',RMSprop(),0.1),
#    (10,10,1,3,75,'elu',Nadam(),0.2)
#]
    
# kodiak edit
best_tuples = [
#    (5,100,1,10,100,'elu',Nadam(),0.2),  # v21 already run
#    (10,10,1,2,75,'elu',Adam(),0.3),  #v22
#    (5,50,1,3,100,'elu',Nadam(),0.1),  #v23
#    (10,100,1,3,75,'elu',Adam(),0.1),  #24
#    (5,100,1,10,100,'relu',Nadam(),0.3),  #v25
#    (10,100,1,3,75,'elu',Adam(),0.2),  # v26
#    (10,10,1,3,100,'elu',RMSprop(),0.1),  # v27
#    (5,50,1,2,75,'elu',Nadam(),0.3),  # v28
#    (10,10,1,3,100,'relu',RMSprop(),0.3), # v29
#    (5,10,1,10,100,'elu',Adam(),0.2),  # v30
#    (10,10,1,3,75,'elu',Adam(),0.3),  # v31
#    (5,10,2,1,100,'elu',Nadam(),0.2),  #v32
#    (5,10,1,1,75,'elu',Nadam(),0.2),  #v33
    # below configs are being run w cudnn verison on workstation
#    (5,10,1,2,100,'elu',RMSprop(),0.2), # already ran this one
#    (10,10,1,3,75,'elu',Adam(),0.2),
    (5,10,1,1,100,'elu',Nadam(),0.2),
    (5,10,1,10,100,'elu',RMSprop(),0.1),
    (10,10,1,3,75,'elu',Nadam(),0.2)
]

additional_params = [
        [0,0.001,0.0001,0.00001],  # min delta for early stop
        [1,0],  # true or false to use learning rate reduction (some optimizers do this anyway)
        [20,50,100],  # patience for early stopping
        [0.5,0.75,0.8],  # factor for lr reduction
        ['val_final_output_mse','val_final_output_loss'],  # quantity to monitor for callbacks
        [1,0]  # use pblf or not
        ]
additional_tuples = list(itertools.product(*additional_params))
# pick up where you left off
#additional_tuples[:662] = []

#%% Call the function to perform hyperparaeter tuning

#for i,tup in enumerate(param_tuples):
#    print(i,tup)
#    params = {
#        'Nnodes_LSTM': tup[0],
#        'Nnodes_MLP': tup[1],
#        'Nlyr_LSTM': tup[2],
#        'Nlyr_MLP': tup[3],
#        'batch_size': tup[4],
#        'MLP_activation': tup[5],
#        'optimizer': tup[6],
#        'do': tup[7]
#        }
#    
#    history,model = LSTM_PGNN(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_shift_test,Ysmap_test,params)
#    
#    histdict = history.history
#    
#    params['optimizer'] = str(params['optimizer'])
#    params['final_output_loss'] = histdict['val_final_output_loss'][len(histdict['val_final_output_loss'])-1]
#    params['final_output_mse'] = histdict['val_final_output_mse'][len(histdict['val_final_output_mse'])-1]
#    
#    param_df = pd.DataFrame(params,index=[0])
#    
#    if not os.path.exists('HPout_cuDNN.csv'):
#        param_df.to_csv('HPout_cuDNN.csv', mode='w+', index=False, header=True)
#    else:
#        param_df.to_csv('HPout_cuDNN.csv', mode='a', index=False, header=False)
#    
##    if not os.path.exists('HPout_cuDNN_kodiak.csv'):
##        param_df.to_csv('HPout_cuDNN_kodiak.csv', mode='w+', index=False, header=True)
##    else:
##        param_df.to_csv('HPout_cuDNN_kodiak.csv', mode='a', index=False, header=False)

# now optimize over additional parameters using the best performers from 6000+ HP runs  
for best in best_tuples: 
    for i,tup in enumerate(additional_tuples):
        tup = best + tup
        print(i,tup)
        params = {
            'Nnodes_LSTM': tup[0],
            'Nnodes_MLP': tup[1],
            'Nlyr_LSTM': tup[2],
            'Nlyr_MLP': tup[3],
            'batch_size': tup[4],
            'MLP_activation': tup[5],
            'optimizer': tup[6],
            'do': tup[7],
            'min_delta': tup[8],  # min delta for early stop
            'TFlr': tup[9],  # true or false to use learning rate reduction (some optimizers do this anyway)
            'patience_es': tup[10],  # patience for early stopping
            'lr_factor': tup[11],  # factor for lr reduction
            'monitor': tup[12],  # quantity to monitor for callbacks
            'TF_pblf': tup[13]  # use pblf or mse
            }
        
        history,model = LSTM_PGNN_addtnl_params(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_test,params)
        
        histdict = history.history
        
        params['optimizer'] = str(params['optimizer'])
        params['final_output_loss'] = histdict['val_final_output_loss'][len(histdict['val_final_output_loss'])-1]
        params['final_output_mse'] = histdict['val_final_output_mse'][len(histdict['val_final_output_mse'])-1]
        
        param_df = pd.DataFrame(params,index=[0])
        
        if not os.path.exists('HPout_additional_cuDNN.csv'):
            param_df.to_csv('HPout_additional_cuDNN.csv', mode='w+', index=False, header=True)
        else:
            param_df.to_csv('HPout_additional_cuDNN.csv', mode='a', index=False, header=False)
        
        












































































