# -*- coding: utf-8 -*-
"""
Read in and preprocess data
"""

# import packages
import os
import pandas as pd
import numpy as np
from sklearn import preprocessing
import h5py
from random import shuffle

# define custom functions
def moving_average(data, window_size):
    window= np.ones(int(window_size))/float(window_size)
    return np.convolve(data, window, 'same')

print('DEFINE WORKING DIRECTORY PATH AS THE VARIABLE "main"')
main = ''
print('STORE DATA FOR ELKR IN THE "main" DIRECTORY')
main = ''
print('DEFINE SWAT FILEPATH AS THE VARIABLE "swat_path"')
print('STORE SWAT FILES FOR ELKR IN THE "swat_path" DIRECTORY')
swat_path = ''

#%% READ

print('Read in data..................')
os.chdir(swat_path)

hru_raw = pd.read_csv('hru-lte.hru',
                         sep='\s+',
                         skiprows=1)
weather = pd.read_csv('wst_hru.out',
                         sep='\s+')
waterbal = pd.read_csv('waterbal_day_sd.txt',
                       sep='\s+',
                       usecols=[0,1,2,16])

# read in ET data
os.chdir(main)
with h5py.File('Y_ET.h5','r') as f:
    Yet = np.array(f['/data']).T


#%% VARIABLES
units = hru_raw['NUMB']
N_samples = hru_raw.shape[0]
padding_value = -1

#%% XS
print('Process Xstatic...................')

# the only non-numeric column is for land use, which uses a four-letter character code. previously, I replaced this with the NLCD numeric identifier equivalent for each land-use class, but that may not have been the best choice because those numbers are relatively arbitrary and do not imply any type of scale (i.e. larger is not 'better'). try using one-hot encoding instead
drop_vars = ['CN3_SWF','TC','PERCO','SY','ABF','REVAPC','PERCC','SW','GW','GWFLOW','GWDEEP','SNOW','TROPICAL','IGROW1','IGROW2','STRESS','IPET','IRR','TDRAIN','USLEP','ETCO','XLAT'] # remove params that have a constant value for all hrus or that just don't make sense
hru = hru_raw.copy()
hru = hru.drop(drop_vars,axis=1)
cat_vars = ['ITEXT','IPLANT']
# when I wrote the orginal data script, I did not anticipate something I should have - the data arrays for the test sites need to be exactly the same number of cols as the training array, which for static data is variable depending on the number of plant and soil textural categories. Fix that by reading in the full US HRU file and finding the unique number of categories for each categorical variable
os.chdir(swat_path)
hru_usa = pd.read_csv('hru-lte_USA.hru',
                         sep='\s+',
                         skiprows=1,
                         usecols=['ITEXT','IPLANT'])
itext_unique = np.unique(hru_usa['ITEXT'])
iplant_unique = np.unique(hru_usa['IPLANT'])
unique_cats = [itext_unique,iplant_unique]
os.chdir(main)

cat_num = [7,15]
for i,var in enumerate(cat_vars):
#    print(i,var)
    tmp = hru[var]
    hru = hru.drop(var,axis=1)
    # get unique categories and the number of times they occur in the data
    new_cols = unique_cats[i]
    for col in new_cols:
#        print(col)
        col = var + '_' + str(col)
        hru[col] = np.zeros(tmp.shape)
    for j,col in enumerate(tmp):
        col = var + '_' + str(col)
        hru.loc[j,col] = 1

Xstatic = hru.iloc[:,2:]
XS_dim = Xstatic.shape

#%% XD

print('Process Xdynamic...................')
N_timesteps = int(weather.shape[0]/N_samples)
XD_N_features = weather.shape[1]-3

XD_dim = (N_samples,N_timesteps,XD_N_features)
Xdyn = np.ones(XD_dim) * padding_value
Xanom = np.ones(XD_dim) * padding_value
Xtremep = np.zeros(XD_dim)
Xtremem = np.zeros(XD_dim)
window_size =90  # calc moving avg using a window of days
stride = int(window_size*.5)  # overlap windows by 50%

for u in units:
    print(u)
    tmp = weather[weather['unit'] == u]
    tmp = tmp.drop(['year','time','unit'],axis=1).reset_index(drop=True)
    Xdyn[u-1,:,:] = tmp
    
    tmp_mu = tmp.copy() * 0
    tmp_sigma = tmp.copy() * 0
    tmp_anom = tmp.copy()
    
    for col in tmp_anom.columns:
        ts = tmp_anom[col]
        ts_mu = np.zeros(ts.shape)
        ts_sigma = np.zeros(ts.shape)
        window_beg_idx = list(range(0, len(ts), stride))
        ts_ma = [np.mean(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        ts_std = [np.std(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        for i,idx in enumerate(window_beg_idx):
#            print(i,i+window_size)
            if idx+window_size <= len(ts):
                tmp_anom.loc[idx:idx+window_size-1,col] = ts[idx:idx+window_size] - ts_ma[i]
                ts_mu[idx:idx+window_size] = ts_ma[i]
                ts_sigma[idx:idx+window_size] = ts_std[i]
            else:
                tmp_anom.loc[idx:len(ts),col] = ts[idx:idx+window_size] - ts_ma[len(ts_ma)-1]
                ts_mu[idx:len(ts)] = ts_ma[len(ts_ma)-1]
                ts_sigma[idx:len(ts)] = ts_std[len(ts_std)-1]
        tmp_mu[col] = ts_mu
        tmp_sigma[col] = ts_sigma
    
    # create additional binary features where a 1 indicates that a value i greater/less than 2sigma above/below the mean (i.e. extreme value)
    tmp_p = np.zeros(tmp.shape)
    tmp_m = np.zeros(tmp.shape)
    
    idx = np.where(tmp >= tmp_mu + tmp_sigma + 0.0001)
    tmp_p[idx[0]-3,idx[1]] = 0.75  # weight antecedent conditions
    tmp_p[idx[0]-2,idx[1]] = 0.75  # weight antecedent conditions
    tmp_p[idx[0]-1,idx[1]] = 0.75  # weight antecedent conditions
    tmp_p[idx] = 1
    
    idx = np.where(tmp <= tmp_mu - tmp_sigma - 0.0001)
    tmp_m[idx[0]-3,idx[1]] = 0.75  # weight antecedent conditions
    tmp_m[idx[0]-2,idx[1]] = 0.75  # weight antecedent conditions
    tmp_m[idx[0]-1,idx[1]] = 0.75  # weight antecedent conditions
    tmp_m[idx] = 1
    
    Xtremep[u-1,:,:] = tmp_p
    Xtremem[u-1,:,:] = tmp_m

    Xanom[u-1,:,:] = tmp_anom

#Xdynamic = np.concatenate((Xdyn,Xanom,Xtremep,Xtremem),axis=2) # SAT FLAGS
Xdynamic = np.concatenate((Xdyn,Xanom),axis=2)  # NO SAT FLAGS
XD_dim = Xdynamic.shape

# write this to file before adding synthetic data
with h5py.File('XDRAW_LSTM.hdf5','w') as f:
    yset = f.create_dataset("XDRAW_LSTM",XD_dim,data=Xdynamic) 
    
#with h5py.File('XDRAW_LSTM.hdf5','r') as f:
#   Xdynamic = np.array(f["XDRAW_LSTM"])
#XD_dim = Xdynamic.shape   


#%% Y_smap

print('Process Ysmap...................')
Ysmap_dim = (N_samples,10000)
Y_smap = np.ones(Ysmap_dim) * np.nan

smap_hru = np.genfromtxt('HRU-SMAP.con',delimiter='\t',skip_header=1,dtype='str')
scan_hru = np.genfromtxt('scan_hru.txt',delimiter='\t',skip_header=1,dtype='str')
tidx = pd.read_csv('tidx_lookup.txt',sep='\t')
tidx['DATETIME'] = pd.to_datetime(tidx['DATETIME'],format='%m/%d/%Y')
smap_time = open('SMAP_timeseries.txt','r')
count = -1
for line in smap_time.readlines():
    count += 1
    if count > 0:
        line = line.split('\t')
        yr = line[0]
        jday = int(line[1]) 
        smid = line[2]
        val = float(line[3])
        if val < -997.0:
             # if there is no SMAP retrieval available, pull in the value from the nearest scan station for any HRU assigned to the SMAP pixel
             smatch = smap_hru[np.where(smap_hru[:,1] == smid),0]
             if smatch.shape[1] == 0:
                 val = np.nan
             else:
                 hru = smatch[0,0]
                 scan = scan_hru[np.where(scan_hru[:,0]==hru),4].item()
                 print('DEFINE A FILEPATH TO SCAN DATA')
                 filepath = ''
                 scandata = pd.read_csv(filepath,
                                          sep=',',
                                          header=0,
                                          comment='#',
                                          usecols=(0,1))
                 scandata['Date'] = pd.to_datetime(scandata['Date'],format='%Y-%m-%d')
                 tmptime = tidx[tidx.YEAR == int(yr)]
                 tmptime = tmptime[tmptime.JDAY == jday]
                 dt = tmptime.DATETIME
                 dtidx = list(dt.index.values)
                 if len(dtidx) > 0:
                     dt = dt[dtidx[0]]
                     scandat = scandata[scandata['Date'] == dt]
                     scancols = list(scandat.columns.values)
                     if scandat.shape[0] > 0:
                         val = scandat[scancols[1]].item()/100
                     else:
                         val = np.nan
                 else:
                     val = np.nan

        samples = smap_hru[np.where(smap_hru[:,1] == smid),0][0]
        tmp1 = tidx[tidx.YEAR == int(yr)]
        tmp2 = tmp1[tmp1.JDAY == int(jday)]
        
        if int(yr) < 2020:
            j = tmp2.T_IDX.item()
            for i in samples:
                Y_smap[int(i)-1,j] = val
    

smap_time.close()

# remove rows of all zeros (extra timesteps)
Y_smap = Y_smap[:,:j+1]
Ysmap_dim = Y_smap.shape

# interpolate over nans (times when neither SMAP or SCAN data are available)
Y_smap_interp = Y_smap.copy()

# first, if endpoints are nans they will not be interpolated, so replace them with the value for the preceding sample
nanval = 0.3
for i,row in enumerate(Y_smap_interp):
    start_idx = 0
    end_idx = Y_smap_interp.shape[1] - 1
    if np.isnan(Y_smap_interp[i,start_idx]):
        if i > 0:
            nanval = Y_smap_interp[i-1,start_idx]
        Y_smap_interp[i,start_idx] = nanval
    if np.isnan(Y_smap_interp[i,end_idx]):
#        print(i)
        if i > 0:
            nanval = Y_smap_interp[i-1,end_idx]
        Y_smap_interp[i,end_idx] = nanval

# now interpolate nans using cubic spline
Y_smap_interp = pd.DataFrame(Y_smap_interp).interpolate(method='cubic',axis=1).values
Y_smap = Y_smap_interp.reshape(Ysmap_dim)

# write this to file before adding synthetic data
with h5py.File('YsmapRAW_LSTM.hdf5','w') as f:
    yset = f.create_dataset("YsmapRAW_LSTM",Ysmap_dim,data=Y_smap)  
    
#with h5py.File('YsmapRAW_LSTM.hdf5','r') as f:
#   Y_smap = np.array(f["YsmapRAW_LSTM"])

#%% normalize inputs

print('normalise inputs.............')
# normalize inputs across samples
Xstatic = preprocessing.normalize(Xstatic,norm='l2',axis=0)
for k in range(Xdynamic.shape[0]):
    # norm='max' indicates use of l-inf norm which would encourage extreme values to not be considered outliers
    Xdynamic[k,:,:] = preprocessing.normalize(Xdynamic[k,:,:],norm='max',axis=1)

# update dimensions
XD_dim = Xdynamic.shape
XS_dim = Xstatic.shape
Ysmap_dim = Y_smap.shape
#%% write matrices for input to DLM

# create "shifted" SM and ET arrays to calculate difference between current and previous day for PBLF
Yet_shift = np.zeros(Yet.shape)
Yet_shift[:,0] = np.random.normal(np.nanmean(Yet),np.nanstd(Yet),Yet_shift[:,0].shape)
Yet_shift[:,1:] = Yet[:,0:Yet.shape[1]-1]

# make sure the phys rel is only used on precip-free days - do this by identifying timesteps with precip and changing the values such that the loss function is not penalized
widx = weather[weather.unit == 1].index
precip = weather.precip[widx]
precip = precip[454:].reset_index(drop=True)  # LIMIT TO ONLY OUTPUT TIMESERIES LENGTH
nzidx = precip[precip != 0].index

# manually make sure that on precip days the loss function is not penalized (that ET and SM changes between day i and i-1 have opposite sign)
Yet[:,nzidx] = Yet[:,nzidx] + 1.0
Yet_shift[:,nzidx] = Yet_shift[:,nzidx] - 1.0
Y_smap[:,nzidx,:] = Y_smap[:,nzidx,:] + 0.1

# make sure to shuffle the rows of all arrays so that when the training/test sets are partitioned they don't have too many hrus with spatial similarity
ismpls = list(i for i in range(0,Xdynamic.shape[0]))
shuffle(ismpls)
ismpls = np.argsort(ismpls)#
Xdynamic = Xdynamic[ismpls,:,:]
Xstatic = Xstatic[ismpls,:]
Y_smap = Y_smap[ismpls,:]
Yet = Yet[ismpls,:]
Yet_shift = Yet_shift[ismpls,:]

# reserve the top Ntest rows for plotting...the first 3200 rows/samples are raw data w/o added noise, which will be needed for plotting and testing later
Ntest = 10
XD_test0 = Xdynamic[:Ntest,:,:]
XS_test0 = Xstatic[:Ntest,:]
Ysmap_test0 = Y_smap[:Ntest,:]
Yet_test0 = Yet[:Ntest,:]
Yet_shift_test0 = Yet_shift[:Ntest,:]

Xdynamic = np.delete(Xdynamic,range(Ntest),0)
Xstatic = np.delete(Xstatic,range(Ntest),0)
Y_smap = np.delete(Y_smap,range(Ntest),0)
Yet = np.delete(Yet,range(Ntest),0)
Yet_shift = np.delete(Yet_shift,range(Ntest),0)

padding_value = -1.0
split_idx = int(round(Xdynamic.shape[0]*0.1))
Y_smap[np.isnan(Y_smap)] = np.random.normal(0.3,0.01,1)
Y_smap[Y_smap<0] = np.random.normal(0.3,0.01,1)
    
XD_Nsamples = Xdynamic.shape[0]
XD_Ntimesteps = Xdynamic.shape[1]
XD_Nfeatures = Xdynamic.shape[2]
XD_test = Xdynamic[:split_idx,:]
XD_train = Xdynamic[split_idx:,:]

XS_Nsamples = Xstatic.shape[0]
XS_Nfeatures = Xstatic.shape[1]
XS_test = Xstatic[:split_idx,:]
XS_train = Xstatic[split_idx:,:]

Ysmap_Nsamples = Y_smap.shape[0]
Ysmap_Nfeatures = Y_smap.shape[1]
Ysmap_test = Y_smap[:split_idx,:]
Ysmap_train = Y_smap[split_idx:,:]

Yet_test = Yet[:split_idx,:]
Yet_train = Yet[split_idx:,:]

Yet_shift_test = Yet_shift[:split_idx,:]
Yet_shift_train = Yet_shift[split_idx:,:]

print('Write h5 files................')

with h5py.File('X_D_LSTM.hdf5','w') as f:
    xDtrain = f.create_dataset("XD_train",XD_train.shape,data=XD_train)
    xDtest = f.create_dataset("XD_test",XD_test.shape,data=XD_test)
        
with h5py.File('X_S_LSTM.hdf5','w') as f:
    xStrain = f.create_dataset("XS_train",XS_train.shape,data=XS_train)
    xStest = f.create_dataset("XS_test",XS_test.shape,data=XS_test)
        
with h5py.File('Y_LSTM.hdf5','w') as f:
    Ysmaptrain = f.create_dataset("Y_train",Ysmap_train.shape,data=Ysmap_train)
    Ysmaptest = f.create_dataset("Y_test",Ysmap_test.shape,data=Ysmap_test)
        
with h5py.File('Y_ET_LSTM.hdf5','w') as f:
    Yettrain = f.create_dataset("Yet_train",Yet_train.shape,data=Yet_train)
    Yettest = f.create_dataset("Yet_test",Yet_test.shape,data=Yet_test)
    Yetshift_test = f.create_dataset("Yet_shift_test",Yet_shift_test.shape,data=Yet_shift_test)
    Yetshift_train = f.create_dataset("Yet_shift_train",Yet_shift_train.shape,data=Yet_shift_train)
        
with h5py.File('Test0.hdf5','w') as f:
    XD_test0 = f.create_dataset('XD_test0',XD_test0.shape,data=XD_test0)
    XS_test0 = f.create_dataset('XS_test0',XS_test0.shape,data=XS_test0)
    Y_test0 = f.create_dataset('Y_test0',Ysmap_test0.shape,data=Ysmap_test0)
    Yet_test0 = f.create_dataset('Yet_test0',Yet_test0.shape,data=Yet_test0)
    Yet_shift_test0 = f.create_dataset('Yet_shift_test0',Yet_shift_test0.shape,data=Yet_shift_test0)

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    